/*
 * Gnophone: A client for the Asterisk PBX
 *
 * Copyright (C) 2000, Linux Support Services, Inc.
 *
 * Written by Mark Spencer
 *
 * Linux/UNIX version distributed under the terms of
 * the GNU General Public License
 *
 * gnophone.c: General initialization code
 *
 */

#include <sched.h>
#include <stdio.h> 
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include "gnophone.h"
#include "audio.h"
#include "telco.h"
#include "htmlengine.h"
#include "phonecore.h"

#ifndef USE_FORK
#include <pthread.h>
#endif

char home_page[256] = DEFAULT_HOME_PAGE;

int is_phone_number(char *s)
{
	char *digits = "()-,0123456789*# ";
	int found = 1;
	
	if (!strlen(s))
		return 0;
	while(found && *s) {
		found = found && strchr(digits, *(s++));
	}
	return found;
}

static int gno_socket = -1;
static int gnoid = -1;
#ifdef USE_FORK
/* PID of phonecore */
static pid_t pid = -1;
#else
static pthread_t thread;
#endif

#ifdef USE_FORK
static void sigchld(int sig)
{
	pid_t p;
	int status;
	p = waitpid(-1, &status, WNOHANG);
	if (p < 0) 
		return;
	if (p == pid) {
		fprintf(stderr, "Phone Core Died!  Please submit a bug report at\n"
						"http://www.gnophone.com/bugs\n"
						"Bug Code: PCDEATH-%d-%d-%d\n", 
							WEXITSTATUS(status), 
							WTERMSIG(status),
							 WSTOPSIG(status));
		exit(1);
	}
	fprintf(stderr, "Huh?  Some child %d died but it wasn't the phone core?\n", p);
}

#endif

static void read_from_socket(void *data, int fd, GdkInputCondition c);
static int make_socket(void)
{
	char filename[80];
	char *h;
	h = getenv("HOME");
	if (!h)
		return 1;
	snprintf(filename, sizeof(filename), "%s/.gnophone", h);
	mkdir(filename, 0700);
	snprintf(filename, sizeof(filename), "%s/.gnophone/comm", h);
	unlink(filename);
	mkfifo(filename, 0700);
	gno_socket = open(filename, O_RDONLY | O_NONBLOCK);
	if (gno_socket > -1) {
		gnoid = gdk_input_add(gno_socket, GDK_INPUT_READ, read_from_socket, NULL);
		return 0;
	}
	return -1;
}

static int connect_socket(void)
{
	char filename[80];
	char *h;
	h = getenv("HOME");
	if (!h)
		return 1;
	snprintf(filename, sizeof(filename), "%s/.gnophone/comm", h);
	gno_socket = open(filename, O_WRONLY | O_NONBLOCK);
	if (gno_socket > -1) {
		return 0;
	}
	return -1;
}

static void read_from_socket(void *data, int fd, GdkInputCondition c)
{
	char buf[256];
	int res;
	gdk_input_remove(gnoid);
	res = read(fd, buf, sizeof(buf) - 1);
	/* Close and reopen */
	close(gno_socket);
	make_socket();
	if (res < 1)
		return;
	buf[res] = '\0';
	if (buf[strlen(buf) - 1] == '\n')
		buf[strlen(buf) - 1] = '\0';
	printf("Remote: %s\n", buf);
	interpret_call(buf);
}

int interpret_call_strict(char *s)
{
	char *proto, *dest, *ihr;
	if (!s || !strlen(s))
		return -1;
	if (!strchr(s, ':'))
		return -1;
	proto = strdup(s);
	dest = proto;
	while(*dest != ':') dest++;
	*dest = '\0';
	dest++;
	if (!strcasecmp(proto, "tel")) {
		fprintf(stderr, "I should dial %s\n", dest);
		ihr = telco_to_irl(dest);
		if (ihr) {
			fprintf(stderr, "Using: %s\n", ihr);
			launch_call(ihr, dest, 0);
		}
	} else if (!strcasecmp(proto, "iax")) {
		if (strncmp(dest, "//", 2)) {
			fprintf(stderr, "Destination '%s' lacks // \n", dest);
			return -1;
		}
		launch_call(dest + 2, dest + 2, 0);
	} else if (!strcasecmp(proto, "dtmf")) {
		while(*dest) {
			digit_ready(*dest, 0, 1);
			dest++;
		}
	} else {
		fprintf(stderr, "Don't know what to do with %s\n", proto);
	}
	free(proto);
	return 0;
		
}

int interpret_call(char *s)
{
	char *ihr;
	char *proto, *dest;
	/* Try to interpret what they type in the space... */
	if (strchr(s, ' ')) {
		/* It's probably a name...  Maybe one day we'll look it
		   up in their address book */
		return -1;
	}
	if (strchr(s, ':')) {
		/* It's either a URL, or an IAX string */
		proto = strdup(s);
		dest = proto;
		while(*dest && (*dest != ':')) dest++;
		*dest = 0;
		dest++;
		if (!strcasecmp(proto, "http") ||
		    !strcasecmp(proto, "https") ||
		    !strcasecmp(proto, "ftp") ||
			!strcasecmp(proto, "mailto") ||
			!strcasecmp(proto, "about") ||
			!strcasecmp(proto, "file") ||
			!strcasecmp(proto, "finger") ||
			!strcasecmp(proto, "keyword")) {
				/* Send it to the primary browser widget */
				start_load_url(s);
				free(proto);
		} else if (!strcasecmp(proto, "iax") && (strlen(dest) > 3)) {
			/* It's an iax:// address */
			free(proto);
			launch_call(s + 6, s + 6, 0);
		} else if (!strcasecmp(proto,"dtmf") && strlen(dest)) {
			while(*dest) {
				digit_ready(*dest, 0, 1);
				dest++;
			}
		} else if (!strcasecmp(proto, "gnophone")) {
			/* An internal gnophone thingy */
			if (!strcasecmp(dest, "audiocontrol"))
				show_audio_control();
			else if (!strcasecmp(dest, "videocontrol"))
				image_callback();
			else if (!strcasecmp(dest, "telephonecontrol"))
				telco_prefs();
		}
	} else {

		if (is_phone_number(s)) {
			printf("%s is a phone number\n", s);
			ihr = telco_to_irl(s);
			if (ihr && strlen(ihr)) {
				printf("Using: %s\n", ihr);
				launch_call(ihr, s, 0);
			} else {
				if (ihr)
					fprintf(stderr, "Couldn't make it: %s\n", ihr + 1);
				return -1;
			}
		} else {
			launch_call(s, s, 0);
		}
		
		/*							*
		 *	Bug #5 , let the widget handle it by default. 	*
		 *	since the window is austensibly for browsing. 	*
		 *	-Mark Holcomb					*/
		/* start_load_url(s); */
	}
	return 0;
}

void interpret_arg(char *arg, int pass)
{
	char buf[256];
	char *c;
	FILE *f = fopen(arg, "r");
	if (f) {
		while(!feof(f)) {
			fgets(buf, sizeof(buf), f);
			if (strlen(buf))
				buf[strlen(buf) - 1] = '\0';
			if ((c = strchr(buf, ';')))
				*c = 0;
			while(strlen(buf) && (buf[strlen(buf)-1] < 33))
				buf[strlen(buf) -1] = '\0';
			if (strlen(buf)) {
				if (pass) {
					if (write(gno_socket, buf, strlen(buf) + 1) != 
						strlen(buf) + 1)
							fprintf(stderr, "Error sending command '%s'\n", buf);
				} else
					interpret_call(buf);
				break;
			}
		}
		fclose(f);
	}
	if (pass) {
		if (write(gno_socket, arg, strlen(arg) + 1) != 
						strlen(arg) + 1) 
				fprintf(stderr, "Error sending command '%s'\n", buf);
	} else
		interpret_call(arg);
}

char *base_filename(char *filename)
{
	static char buf[256];
	char *prefix;
	if (!filename)
		return NULL;
	if (filename[0] == '/')
		return filename;
	if (!(prefix = getenv("GNOPHONE_HOME")))
		prefix=DEFAULT_DIR;
	snprintf(buf, sizeof(buf), "%s/%s", prefix,filename);
	return buf;
}

static int nobrowser=0;
static int quiet;
static int rtt;
static struct option options[] =
{
	{ "nobrowser", 0, &nobrowser, 1 },
	{ "quiet", 0, &quiet, 1 },
	{ "realtime", 0, &rtt, 1 },
	{ NULL, },
};

static void *pc_process(void *unused)
{
	struct sched_param sp;
	if (rtt) {
		memset(&sp, 0, sizeof(sp));
		sp.sched_priority = 20;
#ifdef USE_FORK
		if (sched_setscheduler(0, SCHED_RR, &sp))
#else		
		if (pthread_setschedparam(pthread_self(), SCHED_RR, &sp))
#endif		
			fprintf(stderr, "Unable to set high priority scheduler\n");
	}
	phonecore_thread(unused);
	fprintf(stderr, "Phonecore Failed!?\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	int best;
	int port;
	int res;
	int fd;
	FILE *f;
	char fn[256];
	struct stat s;
	struct audio_channel *ac, *bac;
	char *arg= NULL;
	
	for(;;) {
		res = getopt_long(argc, argv, "", options, NULL);
		if (res < 0)
			break;
		if (res == '?')
			exit(0);
	}
	
	if (quiet) {
		/* XXX Ugh, this is nasty, but hopefully this will
		   get everything quiet, even mozilla's chatter XXX */
		fd = open("/dev/null", O_WRONLY);
		dup2(fd, STDOUT_FILENO);
		dup2(fd, STDERR_FILENO);
		close(fd);
	}
	if (rtt && geteuid()) {
		fprintf(stderr, "Realtime thread requires root priviledges.\n");
		exit(1);
	}
	
	if (optind < argc)
		arg = argv[optind];
	
	if (!connect_socket()) {
		if (arg) 
			interpret_arg(arg, 1);
		close(gno_socket);
		exit(0);
	}

	/* Load modules */
	load_modules();

	telco_load_settings();
	video_load_settings();

	if (make_socket()) {
		fprintf(stderr, "Unable to make socket\n");
		return 1;
	}

	if ((port = pc_init()) < 0)
		return 1;


#ifdef USE_FORK
	signal(SIGCHLD, sigchld);

	pid = fork();
	if (!pid)
		pc_process(NULL);

	if (pid < 0) {
		fprintf(stderr, "Fork failed!\n");
		exit(1);
	}
	
#else
	pthread_create(&thread, NULL, pc_process, NULL);
#endif

	if (nobrowser) {
		html_select_engine(NULL);
	} else if (!engine)
		html_select_engine(engines);
	
	if (!audioc && aclist) {
		ac = aclist;
		bac = aclist;
		best = aclist->priority;
		while(ac) {
			if (ac->priority > best) {
				bac = ac;
				best = ac->priority;
			}
			ac = ac->next;
		}
		audio_select(bac);
	}
	
	set_rec_source();
	if (!audioc) {
		fprintf(stderr, "No audio devices found\n");
		return 1;
	}
	if (init_gphx_subsystem(&argc, &argv)) {
		fprintf(stderr, "Failed to intiialize graphics subsystem\n");
		return 1;
	}
	if (build_gui()) {
		fprintf(stderr, "Failed to build GUI\n");
		return 1;
	}

	/* Here we should check to see if we've run gnophone before.
	* If not, we should show the congratulations page. Otherwise,
	* we should carry them over to the login page */

	if (getenv("HOME"))
		snprintf(fn, sizeof(fn), "%s/.gnophone/%s", getenv("HOME"), GNOPHONE_CONFIG);
	else
		snprintf(fn, sizeof(fn), "%s", GNOPHONE_CONFIG);


	if (!stat(fn, &s)) {
		if (start_load_url(STARTING_URL)) {
			fprintf(stderr, "Failed to load starting URL\n");
			return 1;
		}
	} else {
		if (start_load_url(DEFAULT_FIRST_USE_PAGE)) {
			fprintf(stderr, "Failed to load starting URL\n");
			return 1;
		} else {
			f = fopen(fn, "w");
			fclose(f);
		}
	}

	if (arg)
		interpret_arg(arg, 0);
	printf("Running GUI\n");
	if (run_gui()) {
		fprintf(stderr, "Failed to run GUI\n");
		return 1;
	}
#ifdef USE_FORK	
	kill(pid, SIGTERM);
#endif	
	exit(0);
}
